#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <jpeglib.h>
#include <opt.h>
#include <math.h>
#include <fftw.h>

char * input_file;
char * output_file;
double sigma = -1;

void abort_(const char * s, ...)
{
	va_list args;
	va_start(args, s);
	vfprintf(stderr, s, args);
	fprintf(stderr, "\n");
	va_end(args);
	abort();
}

/* Fouriew transform */

fftw_complex * ftr=NULL;
fftw_complex * ftg=NULL;
fftw_complex * ftb=NULL;

/* we will be using this uninitialized pointer later to store raw, uncompressd image */
JSAMPARRAY row_pointers = NULL;
JSAMPARRAY new_pointers = NULL;


/* dimensions of the image we want to write */
JDIMENSION width;
JDIMENSION height;
int num_components;
int quality = 90;
J_COLOR_SPACE color_space;

/**
 * read_jpeg_file Reads from a jpeg file on disk specified by filename and saves into the 
 * raw_image bufter in an uncompressed format.
 * 
 * \returns positive integer if successful, -1 otherwise
 * \param *filename char string specifying the file name to read from
 *
 */

void read_jpeg_file( char *filename )
{
	/* these are standard libjpeg structures for reading(decompression) */
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	/* libjpeg data structure for storing one row, that is, scanline of an image */
	int x, y;
	
	FILE *infile = fopen( filename, "rb" );
	unsigned long location = 0;
	int i = 0;
	
	if ( !infile )
	{
		abort_("Error opening input jpeg file %s!\n", filename);
	}
	/* here we set up the standard libjpeg error handler */
	cinfo.err = jpeg_std_error( &jerr );
	jpeg_create_decompress( &cinfo );
	/* this makes the library read from infile */
	jpeg_stdio_src( &cinfo, infile );
	/* reading the image header which contains image information */
	jpeg_read_header( &cinfo, TRUE );


	/* Start decompression jpeg here */
	jpeg_start_decompress( &cinfo );
	width = cinfo.output_width;
	height = cinfo.output_height;
	num_components = cinfo.out_color_components;
	color_space = cinfo.out_color_space;
	
	size_t rowbytes = width * num_components;

	/* allocate memory to hold the uncompressed image */
	row_pointers = (JSAMPARRAY) malloc(sizeof(j_common_ptr) * height);
	new_pointers = (JSAMPARRAY) malloc(sizeof(j_common_ptr) * height);
	for (y=0; y<height; y++){
		row_pointers[y] = (JSAMPROW) malloc(rowbytes);
		new_pointers[y] = (JSAMPROW) malloc(rowbytes);
	}
	

	/* read one scan line at a time */
	y=0;
	JSAMPARRAY tmp = row_pointers;
	while( cinfo.output_scanline < cinfo.image_height )
	{
		y = jpeg_read_scanlines( &cinfo, tmp, 1 );
		tmp +=y;
	}
	/* wrap up decompression, destroy objects, free pointers and close open files */
	jpeg_finish_decompress( &cinfo );
	jpeg_destroy_decompress( &cinfo );
// 	free( row_pointer[0] );
	fclose( infile );
	/* yup, we succeeded! */
}


int min(int x, int y){
   return x<y?x:y;
}

void gauss(fftw_real sigma){
  fftwnd_plan plan;
  fftw_real scale;
  fftw_real d2, s2, dd;
  fftw_real X, Y;
  int x, y, j;
  
  scale = (fftw_real)width * (fftw_real)height;
  s2 = M_PI*M_PI*2*sigma*sigma;
  
  plan = fftw2d_create_plan(width, height, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);
  fftwnd_one(plan, ftr, NULL);
  fftwnd_one(plan, ftg, NULL);
  fftwnd_one(plan, ftb, NULL);
  
  fftwnd_destroy_plan(plan);
  
  for (y=0; y<height; y++) {
    Y = (fftw_real)min(y,height-y)/(fftw_real)height;
    for (x=0; x<width; x++) {
        X = (fftw_real)min(x, width-x)/(fftw_real)width;
	d2 = X*X + Y*Y;
	j= y + x*height;
	dd = exp(d2*s2);
	ftr[j].re /= scale*dd;
	ftg[j].re /= scale*dd;
	ftb[j].re /= scale*dd;
	ftr[j].im /= scale*dd;
	ftg[j].im /= scale*dd;
	ftb[j].im /= scale*dd;
    }
  }

  plan = fftw2d_create_plan(width, height, FFTW_BACKWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);
  fftwnd_one(plan, ftr, NULL);
  fftwnd_one(plan, ftg, NULL);
  fftwnd_one(plan, ftb, NULL);
  fftwnd_destroy_plan(plan);
}

JSAMPLE get_byte(fftw_real x){
  JSAMPLE sample = x+0.5;
  if (sample < 0) return 0;
  if (sample > 255) return 255;
  return sample;
}
  

void process_file(){

	int x, y, j;

	/* allocate memory to hold the Fourier trabsform */
	ftr = (fftw_complex *) malloc(sizeof(fftw_complex)*width*height);
	ftg = (fftw_complex *) malloc(sizeof(fftw_complex)*width*height);
	ftb = (fftw_complex *) malloc(sizeof(fftw_complex)*width*height);

	for (y=0; y<height; y++) {
	  JSAMPROW row = row_pointers[y];
	  for (x=0; x<width; x++) {
	      JSAMPROW ptr = &(row[x*3]);
	      j= y + x*height;
	      ftr[j].re = ptr[0];
	      ftr[j].im = 0;
	      ftg[j].re = ptr[1];
	      ftg[j].im = 0;
	      ftb[j].re = ptr[2];
	      ftb[j].im = 0;
	  }
	}

	if(sigma>0){
		gauss(sigma);
	}

	for (y=0; y<height; y++) {
	  JSAMPROW new_row = new_pointers[y];
	  for (x=0; x<width; x++) {
	      JSAMPROW new_ptr = &(new_row[x*3]);
	      j= y + x*height;
	      new_ptr[0] = get_byte(ftr[j].re);
	      new_ptr[1] = get_byte(ftg[j].re);
	      new_ptr[2] = get_byte(ftb[j].re);
	  }
	}
	
	free(ftr);
	free(ftg);
	free(ftb);
}

/**
 * write_jpeg_file Writes the raw image data stored in the raw_image bufter
 * to a jpeg image with default compression and smoothing options in the file
 * specified by *filename.
 *
 * \returns positive integer if successful, -1 otherwise
 * \param *filename char string specifying the file name to save to
 *
 */
void write_jpeg_file( char *filename )
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	int y;
	JSAMPARRAY tmp;

	
	/* this is a pointer to one row of image data */
	FILE *outfile = fopen( filename, "wb" );
	
	if ( !outfile )	{
		abort_("Error opening output jpeg file %s!\n", filename );
	}
	cinfo.err = jpeg_std_error( &jerr );
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo, outfile);

	/* Setting the parameters of the output file here */
	cinfo.image_width = width;	
	cinfo.image_height = height;
	cinfo.input_components = num_components;
	cinfo.in_color_space = color_space;
    /* default compression parameters, we shouldn't be worried about these */
	jpeg_set_defaults( &cinfo );
	jpeg_set_quality (&cinfo, quality, TRUE);
	/* Now do the compression .. */
	jpeg_start_compress( &cinfo, TRUE );
	/* like reading a file, this time write one row at a time */
	tmp = new_pointers;
	while( cinfo.next_scanline < cinfo.image_height )
	{
		y = jpeg_write_scanlines( &cinfo, tmp, 1 );
		tmp +=y;
	}
	/* similar to read file, clean up after we're done compressing */
	jpeg_finish_compress( &cinfo );
	jpeg_destroy_compress( &cinfo );
	fclose( outfile );
	
        /* cleanup heap allocation */
	for (y=0; y<height; y++){
		free(row_pointers[y]);
		free(new_pointers[y]);
	}
	free(row_pointers);
	free(new_pointers);
}




int main(int argc, char **argv)
{
	
  optrega(&input_file, OPT_STRING, 'i', (char*) "input-file", (char*) "Input JPEG File");
  optrega(&output_file, OPT_STRING, 'o', (char*) "output-file", (char*) "Output JPEG File");
  optrega(&sigma, OPT_DOUBLE, 'g', (char*) "gauss", (char*) "Gaussian filter");
  opt(&argc,&argv);
  opt_free();
 printf("sigma: %f\n", sigma);
  
  read_jpeg_file(input_file);
  process_file();
  write_jpeg_file(output_file);
  return 0;
}


